::spawnsDelegate <- {

	setTeam = function(teamNum) {
		filter.SetTeam(teamNum)
		respawnRoom.SetTeam(teamNum)
		local drawFunc = teamNum == 0 ? "DisableDraw" : "EnableDraw"
		foreach (rrv in respawnRoomVisualizers) rrv[drawFunc]()
		this.teamNum = teamNum
		return this
	}

	respawnPlayersWithinBoundaries = function() {
		local setOfPlayersToRespawn = []
		foreach (boundary in boundaries) {
			local boundaryScope = ::getScope(boundary)
			foreach (player in boundaryScope.touchers) {
				if (player.IsValid() && setOfPlayersToRespawn.find(player) == null) setOfPlayersToRespawn.push(player)
			}
			boundaryScope.touchers = []
		}
		foreach (player in setOfPlayersToRespawn) player.ForceRespawn()
		return this
	}

	relocateSpawningPlayers = function() {
		foreach (spawnpoint in spawnpoints) spawnpoint.relocateAssignees()
		return this
	}

	setActive = function(active) {
		this.active = active
		return this
	}

	getBestSpawnpoint = function() {
		return ::pickRandom(::extractPrioritized(
			spawnpoints,
			@(sp) sp.priority
		))
	}

	distanceFrom = function(origin) {
		return proximityRefs
		.map(@(ref) (ref.GetOrigin()-origin).Length())
		.reduce(
			@(prev,next) next < prev || prev == null ? next : prev,
			null
		)
	}

}

::spawnpointsDelegate <- {

	addAssignee = function(player) {
		priority--
		assignees.push(player)
		::getScope(player).nextSpawnpoint = this
		if (assignees.len() == 1) onFilledAssignees()
	}

	removeAssignee = function(player) {
		::safeRemoveFrom(assignees, assignees.find(player))
		::getScope(player).nextSpawnpoint = null
		if (assignees.len() == 0) onEmptyAssignees()
	}

	spawnPlayer = function(player) {
		removeAssignee(player)
		if (player.GetTeam() == spawn.teamNum) {
			player.SetAbsOrigin(entity.GetOrigin())
			player.SnapEyeAngles(entity.GetAbsAngles())
			::call(entity, "handleTeleportedIn", player) // Hook
		}
		else {
			::call(entity, "handleTeleportedInFailure") // Hook
			::assignSpawnpointToPlayer(player)
			local nextSpawnpoint = ::getScope(player).nextSpawnpoint
			if (nextSpawnpoint) nextSpawnpoint.spawnPlayer(player)
		}
	}

	relocateAssignees = function() {
		local players = assignees
		assignees = []
		onEmptyAssignees()
		foreach (player in players) {
			::assignSpawnpointToPlayer(player)
		}
	}

	onFilledAssignees = function() {
		removeInvalidAssignees()
		::call(entity, "handleTeleportingInStart") // Hook
	}

	onEmptyAssignees = function() {
		::call(entity, "handleTeleportingInEnd") // Hook
	}

	removeInvalidAssignees = function(firstIteration = true) {
		if (firstIteration && invalidRemoverIsRunning) return
		invalidRemoverIsRunning = true
		local removed = ::safeRemoveFrom(assignees, ::firstIn(assignees, @(assignee) !assignee.IsValid()))
		if (removed) ::call(entity, "handleTeleportedInFailure") // Hook
		if (assignees.len() == 0) {
			invalidRemoverIsRunning = false
			onEmptyAssignees()
		}
		else {
			local delay = ::call(entity, "getFailureCheckDelay") || {value = RandomFloat(5.0, 6.0)} // Hook
			::createTimer(delay.value, @[this]() removeInvalidAssignees(false))
		}
	}

}

::spawn <- {}
::setupSpawn <- function(group, teamNum = 0, active = true) {

	// To be returned
	local spawn = {
		group = group
		teamNum = teamNum
		active = active
		priority = 99999
	}.setdelegate(::spawnsDelegate)
	::spawn[group] <- spawn

	// Create the filter used by the associated door triggers
	spawn.filter <- SpawnEntityFromTable("filter_activator_tfteam", {})

	// Setup door triggers
	foreach (doorTrigger in ::collect("door_trigger_"+group)) {
		NetProps.SetPropEntity(doorTrigger, "m_hFilter", spawn.filter)
	}

	// Setup respawn room
	local respawnRoomName = "respawn_room_"+group
	spawn.respawnRoom <- ::first(respawnRoomName)
	EntityOutputs.AddOutput(spawn.respawnRoom, "OnStartTouch", "!activator", "RunScriptCode", "respawnReferenceOrigin = self.GetOrigin()", 0, -1)

	// Get the respawn room visualizers, no targetnames, just look for those that reference the respawnroom
	spawn.respawnRoomVisualizers <- ::collect("func_respawnroomvisualizer", "FindByClassname")
	.filter((@(_,rrv) NetProps.GetPropString(rrv,"m_iszRespawnRoomName") == respawnRoomName))

	// Get the boundaries in which players should be forced to respawn when the team changes
	spawn.boundaries <- ::collect("boundary_"+group)
	foreach (boundary in spawn.boundaries) {
		::getScope(boundary).touchers <- []
		EntityOutputs.AddOutput(boundary, "OnStartTouch", "!self", "RunScriptCode", "touchers.push(activator)", 0, -1)
		EntityOutputs.AddOutput(boundary, "OnEndTouch", "!self", "RunScriptCode", "touchers = touchers.filter(@(_,toucher) toucher != activator && toucher.IsValid())", 0, -1)
	}

	// Get spawnpoints
	spawn.spawnpoints <- ::collect("spawnpoint_"+group).apply(function(sp) {
		::getScope(sp).spawn <- spawn
		return {
			spawn = spawn
			entity = sp,
			priority = 99999
			assignees = []
			invalidRemoverIsRunning = false
		}.setdelegate(::spawnpointsDelegate)
	})

	// proximity references
	spawn.proximityRefs <- ::collect("proximity_ref_"+group)

	return spawn.setTeam(teamNum)
	
}

// Assign a player to its most suitable spawnpoint
::assignSpawnpointToPlayer <- function(player) {
	local availableSpawns = ::spawn.filter(@(_,spawn) spawn.active && player.GetTeam() == spawn.teamNum)
	local normalSpawns = []
	local proximitySpawns = []
	foreach (_,spawn in availableSpawns) {
		(spawn.proximityRefs.len() == 0 ? normalSpawns : proximitySpawns).push(spawn)
	}
	local spawnChoices, decreasePriority = false
	if (::getScope(player).respawnReferenceOrigin) {
		local closestProximitySpawn = ::pickRandom(::extractPrioritized(proximitySpawns, @(spawn) -spawn.distanceFrom(::getScope(player).respawnReferenceOrigin)))
		spawnChoices = closestProximitySpawn ? normalSpawns.append(closestProximitySpawn) : normalSpawns
	}
	else {
		spawnChoices = normalSpawns.extend(proximitySpawns)
		decreasePriority = true
	} 
	local bestSpawn = ::pickRandom(::extractPrioritized(spawnChoices, @(s) s.priority))
	if (!bestSpawn) return ::printError("Failed to find a spawn for a "+::team(player.GetTeam())+" player! Make sure to have at least one active spawn for each team at all times.")
	local bestSpawnpoint = bestSpawn.getBestSpawnpoint()
	if (decreasePriority) bestSpawn.priority--
	if (!bestSpawnpoint) return ::printError("Failed to find a spawnpoint in the \""+bestSpawn.group+"\" spawn! Make sure to have at least one entity called \"spawnpoint_"+bestSpawn.group+"\".")
	bestSpawnpoint.addAssignee(player)
}

// Adds the required members to the player (and resets them)
::setupPlayerSpawnLogic <- function(player) {
	local playerScope = ::getScope(player)
	playerScope.respawnReferenceOrigin <- null
	playerScope.nextSpawnpoint <- null
}

// Create the trigger that teleports the players that just respawned to their spawnpoint
::spawnPlayer <- function(player) {
	local playerScope = ::getScope(player)
	if (!playerScope.nextSpawnpoint) ::assignSpawnpointToPlayer(player)
	if (!playerScope.nextSpawnpoint) return ::printError("Failed to spawn a player!")
	playerScope.nextSpawnpoint.spawnPlayer(player)
}
::setupSpawning <- function() {
	::createTimer(
		0,
		function() {
			// Using a trigger to handle the eureka effect
			::spawnTrigger(
				::first("info_player_teamspawn", "FindByClassname").GetOrigin(),
				Vector(-25,-25,0),
				Vector(25,25,84),
				::spawnPlayer
			)
		}
	)
}

// Setup players that just joined
::getEvent("player_spawn").addEffect(
	function(data) {
		local player = GetPlayerFromUserID(data.userid)
		// Fires only when the player enters the game
		if (player.GetTeam() == 0) return ::setupPlayerSpawnLogic(player)
	}
)

// Assign a spawnpoint to spawn at to players that just died
::getEvent("player_death").addEffect(
	function(data) {
		local player = GetPlayerFromUserID(data.userid)
		::getScope(player).respawnReferenceOrigin = player.GetOrigin()
		::assignSpawnpointToPlayer(player)
	}
)

// End of round event to reset players' scopes
::getEvent("stats_resetround").addEffect(
	function(data) {
		foreach (player in ::collect("player", "FindByClassname")) ::setupPlayerSpawnLogic(player)
	}
)