::pickRandom <- function(array) {
	if (array.len() == 0) return null
	return array[RandomInt(0,array.len()-1)]
}

::shuffle <- function(array) {
	local shuffledArray = []
	while (array.len() > 0) shuffledArray.push(array.remove(RandomInt(0,array.len()-1)))
	return shuffledArray;
}

::extractPrioritized <- function(entities, priorityFunc) {
	local extracted = entities.reduce(
		function (prev, next) {
			local nextPriority = priorityFunc(next)
			if (prev == null || nextPriority > prev.priority) return {
				priority = nextPriority,
				list = [next]
			}
			if (nextPriority < prev.priority) return prev
			prev.list.push(next)
			return prev
		},
		null
	)
	return extracted ? extracted.list : []
}

::firstIn <- function(array, condition) {
	foreach (idx,item in array) {
		if (condition(item)) return idx
	}
	return null
}

::safeRemoveFrom <- function(array, idx) {
	if (idx == null) return null
	return array.remove(idx)
}

::groupByDelegate <- {_get = @(idx) []}
::groupBy <- function(array, getGroup) {
	local table = {}.setdelegate(::groupByDelegate)
	foreach (item in array) {
		local group = getGroup(item)
		if (group in table) table[group].push(item)
		else table[group] <- [item] 
	}
	return table
}